/*
* Usman Dastgeer, PELAB, IDA
* Multicore Computing Course, Sep 2013.
* A simple Map skeleton example for an integer array (can be made generic using e.g. C++ templates) in OpenCL.
* The code is just for demonstration and is not well-written (e.g. no proper usage of const keyword, no systematic error-checking etc.)
* Compile command: "g++ opencl_map_ex.cpp -L/usr -lOpenCL"
*/

#include <CL/cl.h>
#include <iostream>
#include <stdio.h>
#include <string.h>


// Skeleton function, could be done in a better way (e.g. separate file)
struct MapFunc
{
   std::string func_CL;
   MapFunc()
   {
      func_CL.append("int opencl_func(int a1, int a2)\n"\
      "{"\
      "return a1 + a2; \n"\
      "}");
   }
};

/* 
* Description: kernel, a nicer way would be probably to have OpenCL code in a separate file.
* Hint: To templatize kernel/map on data type, simple write <TYPE> instead of int and then replace this at runtime
* using string replace with the actual type, same for above string OpenCL function.
*/
static std::string MapKernel_CL(
"__kernel void MapKernel_CL(__global int* input1,__global int* input2, __global int* output, unsigned int numElements)\n"
"{\n"
"    int i = get_global_id(0);\n"
"    if(i < numElements)\n"
"    {\n"
"        output[i] = opencl_func(input1[i],input2[i]);\n"
"    }\n"
"}\n"
);


// OpenCL globals for simplicity
cl_kernel kernel;
cl_device_id deviceID;
cl_context context;
cl_command_queue queue;


// function that do the initialization, e.g. creating context, queue and doing kernel compilation
void initializeOpenCL(MapFunc mfunc)
{
   cl_int err;
   cl_platform_id temp_platforms[2]; // if you have more that 1 OpenCL platform
   
   cl_uint no_platforms; // will contain actual platforms found
   
   // return actual platforms found with along their count
   err = clGetPlatformIDs(2, temp_platforms, &no_platforms);
   if(err != CL_SUCCESS)
   {
      std::cerr<<"Error getting platforms list!!\n"; 
      return;
   }
   
   // we try to select the NVIDIA OpenCL platform, in this case
   int platform_ind=0;
   for(; platform_ind <no_platforms; platform_ind++)
   {
      char inf[1024];
      cl_platform_info temp_info;
      err= clGetPlatformInfo(temp_platforms[platform_ind], CL_PLATFORM_VENDOR, sizeof(char)*1024, inf, NULL);

      if(!strcmp(inf, "NVIDIA Corporation"))
      {
         break; 
      }
   }
   
   // Get OpenCL enabled GPU devices only, we are not interested in CPUs for this lab. 
   cl_uint numDevices;
   clGetDeviceIDs(temp_platforms[platform_ind], CL_DEVICE_TYPE_GPU, 0, NULL, &numDevices);

   if(numDevices == 0)
   {
      std::cerr<<"No OpenCL devices found!\n";
   }

   //Specify platform properties
   cl_context_properties props[3];
   props[0] = (cl_context_properties)CL_CONTEXT_PLATFORM;
   props[1] = (cl_context_properties)temp_platforms[platform_ind];
   props[2] = (cl_context_properties)0;

   // Get device id for a OpenCL GPU device to run out program on that device
   err = clGetDeviceIDs(temp_platforms[platform_ind], CL_DEVICE_TYPE_GPU, 1, &deviceID, NULL);
   if(err != CL_SUCCESS)
   std::cerr<<"Error getting device!!\n";
   
   // Create an OpenCL context, A OpenCL context models all the resources
   context = clCreateContext(0, 1, &deviceID, NULL, NULL, &err);
   if(err != CL_SUCCESS)
   std::cerr<<"Error creating context!!\n" <<err <<"\n";
   
   // Create a queue to execute OpenCL commands
   queue = clCreateCommandQueue(context, deviceID, 0, &err);
   if(err != CL_SUCCESS)
   std::cerr<<"Error creating queue!!\n" <<err <<"\n";

   // Program compilation code...    
   std::string totalSource;
   
   // Create OpenCL source, can be loaded from files (more elegant probably)
   totalSource.append(mfunc.func_CL);
   totalSource.append(MapKernel_CL);
   
   const char* c_src = totalSource.c_str();
   cl_program temp_program;

   // Create program from source
   temp_program = clCreateProgramWithSource(context, 1, &c_src, NULL, &err);
   if(err != CL_SUCCESS){std::cerr<<"Error creating OpenCLsource!!\n";}


   // Build the program (can contain >1 kernels and other OpenCL code as well)
   err = clBuildProgram(temp_program, 0, NULL, NULL, NULL, NULL);
   if (err == CL_BUILD_PROGRAM_FAILURE) 
   {
      // Determine the size of the log
      size_t log_size;
      clGetProgramBuildInfo(temp_program, deviceID, CL_PROGRAM_BUILD_LOG, 0, NULL, &log_size);

      // Allocate memory for the log
      char *log = (char *) malloc(log_size);

      // Get the log
      clGetProgramBuildInfo(temp_program, deviceID, CL_PROGRAM_BUILD_LOG, log_size, log, NULL);

      // Print the log
      std::cerr << log << "\n";
   }
   else if(err != CL_SUCCESS){ std::cerr<<"Error building OpenCL program!!\n" <<err <<"\n"; }

   // Create the OpenCL kernel, best option is to reuse this kernel many times 
   kernel = clCreateKernel(temp_program, "MapKernel_CL", &err);
   if(err != CL_SUCCESS){std::cerr<<"Error creating a kernel!!\n" <<err <<"\n";}
}

// Skeleton implementation, can make it generic easily by using c++ template
void map_func(int *input1, int *input2, int *output, const int size, MapFunc mfunc)
{
   // Do the initialization, better to call it somewhere else(e.g. in main)
   initializeOpenCL(mfunc);
   
   cl_int err;
   
   // It's not portable to hard code threadsize as different machines may have different limit for maximum threadsize that you can have
   // which you can know by quering device properties but here we keep it simple.
   size_t numThreads = 256;
   size_t numBlocks = (size+numThreads-1)/numThreads;
   size_t globalWorkSize[1];
   size_t localWorkSize[1];
   
   size_t sizeVec = size*sizeof(int);
   
   // Allocate memory for dev_in1 and copy from host to device
   cl_mem dev_in1;
   dev_in1 = clCreateBuffer(context, CL_MEM_READ_ONLY, sizeVec, NULL, &err);
   if(err != CL_SUCCESS)
   std::cerr<<"Error allocating memory on device\n";
   err = clEnqueueWriteBuffer(queue, dev_in1, CL_TRUE, 0, sizeVec, (void*)input1, 0, NULL, NULL);
   if(err != CL_SUCCESS)
   std::cerr<<"Error copying data to device\n";
   
   // Allocate memory for dev_in2 and copy from host to device	
   cl_mem dev_in2;
   dev_in2 = clCreateBuffer(context, CL_MEM_READ_ONLY, sizeVec, NULL, &err);
   if(err != CL_SUCCESS)
   std::cerr<<"Error allocating memory on device\n";
   err = clEnqueueWriteBuffer(queue, dev_in2, CL_TRUE, 0, sizeVec, (void*)input2, 0, NULL, NULL);
   if(err != CL_SUCCESS)
   std::cerr<<"Error copying data to device\n";
   
   // Allocate memory for dev_out and copy from host to device	
   cl_mem dev_out;
   dev_out = clCreateBuffer(context, CL_MEM_WRITE_ONLY, sizeVec, NULL, &err);
   if(err != CL_SUCCESS)
   std::cerr<<"Error allocating memory on device\n";
   
   
   // Set kernel argument to call kernel
   clSetKernelArg(kernel, 0, sizeof(cl_mem), (void*)&dev_in1);
   clSetKernelArg(kernel, 1, sizeof(cl_mem), (void*)&dev_in2);
   clSetKernelArg(kernel, 2, sizeof(cl_mem), (void*)&dev_out);
   clSetKernelArg(kernel, 3, sizeof(unsigned int), (void*)&size);

   globalWorkSize[0] = numBlocks * numThreads;
   localWorkSize[0] = numThreads;
   
   // Calling the kernel asynchronously
   err = clEnqueueNDRangeKernel(queue, kernel, 1, NULL, globalWorkSize, localWorkSize, 0, NULL, NULL);
   if(err != CL_SUCCESS){std::cerr<<"Error launching kernel!!\n";}
   
   // Blocking for kernel execution
   clFinish(queue);
   
   // Writing data back form device to host
   err = clEnqueueReadBuffer(queue, dev_out, CL_TRUE, 0, sizeVec, (void*)output, 0, NULL, NULL);
   if(err != CL_SUCCESS)
   std::cerr<<"Error copying from device to host!!\n";
   
   // Releasing resources...
   clReleaseCommandQueue(queue);
   clReleaseContext(context);
}

int main()
{
   const int SIZE=2<<14;
   int input1[SIZE], input2[SIZE], output[SIZE];
   
   for(int i=0;i<SIZE;i++)
   {
      input1[i]=2;
      input2[i]=5;
      output[i]=0;
   }
   
   MapFunc mapFun;
   map_func(input1, input2, output, SIZE, mapFun);
   
   int total=0;
   for(int i=0;i<SIZE;i++)
   {
      total += output[i];
   }	
   
   std::cout<<"Total: "<<total<<"\n";
   // OUTPUT: Total: 57344
}